@@ -2,13 +2,11 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 // Patch file for the dart:async library.
 
+import 'dart:_internal' show patch;
 import 'dart:_js_helper'
     show
-        patch,
         ExceptionAndStackTrace,
         convertDartClosureToJS,
         getTraceFromException,
@@ -20,8 +18,6 @@ import 'dart:_foreign_helper' show JS, JS_GET_FLAG;
 
 import 'dart:_async_await_error_codes' as async_error_codes;
 
-import "dart:collection" show IterableBase;
-
 @patch
 class _AsyncRun {
   @patch
@@ -43,12 +39,12 @@ class _AsyncRun {
       // Use mutationObservers.
       var div = JS('', 'self.document.createElement("div")');
       var span = JS('', 'self.document.createElement("span")');
-      var storedCallback;
+      void Function()? storedCallback;
 
       internalCallback(_) {
         var f = storedCallback;
         storedCallback = null;
-        f();
+        f!();
       }
 
       var observer = JS('', 'new self.MutationObserver(#)',
@@ -93,15 +89,6 @@ class _AsyncRun {
   }
 }
 
-@patch
-class DeferredLibrary {
-  @patch
-  Future<Null> load() {
-    throw 'DeferredLibrary not supported. '
-        'please use the `import "lib.dart" deferred as lib` syntax.';
-  }
-}
-
 @patch
 class Timer {
   @patch
@@ -122,7 +109,7 @@ class Timer {
 
 class _TimerImpl implements Timer {
   final bool _once;
-  int _handle;
+  int? _handle;
   int _tick = 0;
 
   _TimerImpl(int milliseconds, void callback()) : _once = true {
@@ -150,7 +137,8 @@ class _TimerImpl implements Timer {
           convertDartClosureToJS(() {
             int tick = this._tick + 1;
             if (milliseconds > 0) {
-              int duration = JS('int', 'Date.now()') - start;
+              int end = JS('int', 'Date.now()');
+              int duration = end - start;
               if (duration > (tick + 1) * milliseconds) {
                 tick = duration ~/ milliseconds;
               }
@@ -197,15 +185,25 @@ class _AsyncAwaitCompleter<T> implements Completer<T> {
 
   _AsyncAwaitCompleter() : isSync = false;
 
-  void complete([FutureOr<T> value]) {
-    if (!isSync || value is Future<T>) {
+  void complete([FutureOr<T>? value]) {
+    // All paths require that if value is null, null as T succeeds.
+    value = (value == null) ? value as T : value;
+    if (!isSync) {
       _future._asyncComplete(value);
+    } else if (value is Future<T>) {
+      assert(!_future._isComplete);
+      _future._chainFuture(value);
     } else {
-      _future._completeWithValue(value);
+      // TODO(40014): Remove cast when type promotion works.
+      // This would normally be `as T` but we use `as dynamic` to make the
+      // unneeded check be implicit to match dart2js unsound optimizations in
+      // the user code.
+      _future._completeWithValue(value as dynamic);
     }
   }
 
-  void completeError(e, [st]) {
+  void completeError(Object e, [StackTrace? st]) {
+    st ??= AsyncError.defaultStackTrace(e);
     if (isSync) {
       _future._completeError(e, st);
     } else {
@@ -220,6 +218,7 @@ class _AsyncAwaitCompleter<T> implements Completer<T> {
 /// Creates a Completer for an `async` function.
 ///
 /// Used as part of the runtime support for the async/await transformation.
+@pragma('dart2js:assumeDynamic') // Global type inference can't see call site.
 Completer<T> _makeAsyncAwaitCompleter<T>() {
   return new _AsyncAwaitCompleter<T>();
 }
@@ -279,7 +278,7 @@ dynamic _asyncRethrow(dynamic object, Completer completer) {
 /// The [bodyFunction] argument is the continuation that should be invoked
 /// when the future completes.
 void _awaitOnObject(object, _WrappedAsyncBody bodyFunction) {
-  Function thenCallback =
+  FutureOr<dynamic> Function(dynamic) thenCallback =
       (result) => bodyFunction(async_error_codes.SUCCESS, result);
 
   Function errorCallback = (dynamic error, StackTrace stackTrace) {
@@ -298,7 +297,7 @@ void _awaitOnObject(object, _WrappedAsyncBody bodyFunction) {
     _Future future = new _Future().._setValue(object);
     // We can skip the zone registration, since the bodyFunction is already
     // registered (see [_wrapJsFunctionForAsync]).
-    future._thenAwait(thenCallback, null);
+    future._thenAwait(thenCallback, errorCallback);
   }
 }
 
@@ -376,7 +375,7 @@ void _asyncStarHelper(
   if (identical(bodyFunctionOrErrorCode, async_error_codes.SUCCESS)) {
     // This happens on return from the async* function.
     if (controller.isCanceled) {
-      controller.cancelationFuture._completeWithValue(null);
+      controller.cancelationFuture!._completeWithValue(null);
     } else {
       controller.close();
     }
@@ -384,7 +383,7 @@ void _asyncStarHelper(
   } else if (identical(bodyFunctionOrErrorCode, async_error_codes.ERROR)) {
     // The error is a js-error.
     if (controller.isCanceled) {
-      controller.cancelationFuture._completeError(
+      controller.cancelationFuture!._completeError(
           unwrapException(object), getTraceFromException(object));
     } else {
       controller.addError(
@@ -394,9 +393,10 @@ void _asyncStarHelper(
     return;
   }
 
+  _WrappedAsyncBody bodyFunction = bodyFunctionOrErrorCode;
   if (object is _IterationMarker) {
     if (controller.isCanceled) {
-      bodyFunctionOrErrorCode(async_error_codes.STREAM_WAS_CANCELED, null);
+      bodyFunction(async_error_codes.STREAM_WAS_CANCELED, null);
       return;
     }
     if (object.state == _IterationMarker.YIELD_SINGLE) {
@@ -410,7 +410,7 @@ void _asyncStarHelper(
           controller.isSuspended = true;
           return;
         }
-        bodyFunctionOrErrorCode(null, async_error_codes.SUCCESS);
+        bodyFunction(async_error_codes.SUCCESS, null);
       });
       return;
     } else if (object.state == _IterationMarker.YIELD_STAR) {
@@ -426,13 +426,13 @@ void _asyncStarHelper(
         int errorCode = controller.isCanceled
             ? async_error_codes.STREAM_WAS_CANCELED
             : async_error_codes.SUCCESS;
-        bodyFunctionOrErrorCode(errorCode, null);
+        bodyFunction(errorCode, null);
       });
       return;
     }
   }
 
-  _awaitOnObject(object, bodyFunctionOrErrorCode);
+  _awaitOnObject(object, bodyFunction);
 }
 
 Stream _streamOfController(_AsyncStarStreamController controller) {
@@ -450,7 +450,7 @@ Stream _streamOfController(_AsyncStarStreamController controller) {
 /// If yielding while the subscription is paused it will become suspended. And
 /// only resume after the subscription is resumed or canceled.
 class _AsyncStarStreamController<T> {
-  StreamController<T> controller;
+  late StreamController<T> controller;
   Stream get stream => controller.stream;
 
   /// True when the async* function has yielded while being paused.
@@ -460,7 +460,7 @@ class _AsyncStarStreamController<T> {
 
   bool get isPaused => controller.isPaused;
 
-  _Future cancelationFuture = null;
+  _Future? cancelationFuture = null;
 
   /// True after the StreamSubscription has been cancelled.
   /// When this is true, errors thrown from the async* body should go to the
@@ -470,7 +470,7 @@ class _AsyncStarStreamController<T> {
 
   add(event) => controller.add(event);
 
-  addStream(Stream<T> stream) {
+  Future addStream(Stream<T> stream) {
     return controller.addStream(stream, cancelOnError: false);
   }
 
@@ -515,6 +515,7 @@ class _AsyncStarStreamController<T> {
 /// Creates a stream controller for an `async*` function.
 ///
 /// Used as part of the runtime support for the async/await transformation.
+@pragma('dart2js:assumeDynamic') // Global type inference can't see call site.
 _makeAsyncStarStreamController<T>(_WrappedAsyncBody body) {
   return new _AsyncStarStreamController<T>(body);
 }
@@ -564,22 +565,21 @@ class _SyncStarIterator<T> implements Iterator<T> {
   dynamic _body;
 
   // The current value, unless iterating a non-sync* nested iterator.
-  T _current = null;
+  T? _current = null;
 
   // This is the nested iterator when iterating a yield* of a non-sync iterator.
-  // TODO(32956): In strong-mode, yield* takes an Iterable<T> (possibly checked
-  // with an implicit downcast), so change type to Iterator<T>.
-  Iterator _nestedIterator = null;
+  Iterator<T>? _nestedIterator = null;
 
   // Stack of suspended state machines when iterating a yield* of a sync*
   // iterator.
-  List _suspendedBodies = null;
+  List? _suspendedBodies = null;
 
   _SyncStarIterator(this._body);
 
   T get current {
-    if (_nestedIterator == null) return _current;
-    return _nestedIterator.current;
+    var nested = _nestedIterator;
+    if (nested == null) return _current as dynamic; // implicit: as T;
+    return nested.current;
   }
 
   _runBody() {
@@ -610,7 +610,7 @@ class _SyncStarIterator<T> implements Iterator<T> {
   bool moveNext() {
     while (true) {
       if (_nestedIterator != null) {
-        if (_nestedIterator.moveNext()) {
+        if (_nestedIterator!.moveNext()) {
           return true;
         } else {
           _nestedIterator = null;
@@ -620,13 +620,14 @@ class _SyncStarIterator<T> implements Iterator<T> {
       if (value is _IterationMarker) {
         int state = value.state;
         if (state == _IterationMarker.ITERATION_ENDED) {
-          if (_suspendedBodies == null || _suspendedBodies.isEmpty) {
+          var suspendedBodies = _suspendedBodies;
+          if (suspendedBodies == null || suspendedBodies.isEmpty) {
             _current = null;
             // Rely on [_body] to repeatedly return `ITERATION_ENDED`.
             return false;
           }
           // Resume the innermost suspended iterator.
-          _body = _suspendedBodies.removeLast();
+          _body = suspendedBodies.removeLast();
           continue;
         } else if (state == _IterationMarker.UNCAUGHT_ERROR) {
           // Rely on [_body] to repeatedly return `UNCAUGHT_ERROR`.
@@ -635,8 +636,15 @@ class _SyncStarIterator<T> implements Iterator<T> {
           JS('', 'throw #', value.value);
         } else {
           assert(state == _IterationMarker.YIELD_STAR);
-          Iterator inner = value.value.iterator;
+          // Casting to avoid the `.iterator` dynamic call would impact the
+          // performance of yield*.
+          // ignore: avoid_dynamic_calls
+          Iterator<T> inner = value.value.iterator;
           if (inner is _SyncStarIterator) {
+            // The test needs to be 'is _SyncStarIterator<T>' for promotion to
+            // work. However, that test is much more expensive, so we use an
+            // unsafe cast.
+            _SyncStarIterator<T> innerSyncStarIterator = JS('', '#', inner);
             // Suspend the current state machine and start acting on behalf of
             // the nested state machine.
             //
@@ -644,7 +652,7 @@ class _SyncStarIterator<T> implements Iterator<T> {
             // suspending the current body when all it will do is step without
             // effect to ITERATION_ENDED.
             (_suspendedBodies ??= []).add(_body);
-            _body = inner._body;
+            _body = innerSyncStarIterator._body;
             continue;
           } else {
             _nestedIterator = inner;
@@ -668,6 +676,7 @@ class _SyncStarIterator<T> implements Iterator<T> {
 /// Creates an Iterable for a `sync*` function.
 ///
 /// Used as part of the runtime support for the async/await transformation.
+@pragma('dart2js:assumeDynamic') // Global type inference can't see call site.
 _SyncStarIterable<T> _makeSyncStarIterable<T>(body) {
   return new _SyncStarIterable<T>(body);
 }
@@ -675,7 +684,7 @@ _SyncStarIterable<T> _makeSyncStarIterable<T>(body) {
 /// An Iterable corresponding to a sync* method.
 ///
 /// Each invocation of a sync* method will return a new instance of this class.
-class _SyncStarIterable<T> extends IterableBase<T> {
+class _SyncStarIterable<T> extends Iterable<T> {
   // This is a function that will return a helper function that does the
   // iteration of the sync*.
   //
@@ -687,10 +696,3 @@ class _SyncStarIterable<T> extends IterableBase<T> {
   Iterator<T> get iterator =>
       new _SyncStarIterator<T>(JS('', '#()', _outerHelper));
 }
-
-@patch
-void _rethrow(Object error, StackTrace stackTrace) {
-  error = wrapException(error);
-  JS('void', '#.stack = #', error, stackTrace.toString());
-  JS('void', 'throw #', error);
-}
