@@ -2,9 +2,6 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
-import 'dart:_js_helper' show patch;
 import 'dart:_internal' hide Symbol;
 import 'dart:async';
 import 'dart:convert';
@@ -75,7 +72,7 @@ class _AsyncDirectoryListerOps {
 @patch
 class _EventHandler {
   @patch
-  static void _sendData(Object sender, SendPort sendPort, int data) {
+  static void _sendData(Object? sender, SendPort sendPort, int data) {
     throw new UnsupportedError("EventHandler._sendData");
   }
 }
@@ -115,7 +112,7 @@ class _File {
   }
 
   @patch
-  static _create(_Namespace namespace, Uint8List path) {
+  static _create(_Namespace namespace, Uint8List path, bool exclusive) {
     throw new UnsupportedError("File._create");
   }
 
@@ -124,6 +121,11 @@ class _File {
     throw new UnsupportedError("File._createLink");
   }
 
+  @patch
+  static List<dynamic> _createPipe(_Namespace namespace) {
+    throw UnsupportedError("File._createPipe");
+  }
+
   @patch
   static _linkTarget(_Namespace namespace, Uint8List path) {
     throw new UnsupportedError("File._linkTarget");
@@ -300,7 +302,7 @@ class _Platform {
 @patch
 class _ProcessUtils {
   @patch
-  static void _exit(int status) {
+  static Never _exit(int status) {
     throw new UnsupportedError("ProcessUtils._exit");
   }
 
@@ -320,7 +322,7 @@ class _ProcessUtils {
   }
 
   @patch
-  static int _pid(Process process) {
+  static int _pid(Process? process) {
     throw new UnsupportedError("ProcessUtils._pid");
   }
 
@@ -347,33 +349,33 @@ class ProcessInfo {
 class Process {
   @patch
   static Future<Process> start(String executable, List<String> arguments,
-      {String workingDirectory,
-      Map<String, String> environment,
-      bool includeParentEnvironment: true,
-      bool runInShell: false,
-      ProcessStartMode mode: ProcessStartMode.normal}) {
+      {String? workingDirectory,
+      Map<String, String>? environment,
+      bool includeParentEnvironment = true,
+      bool runInShell = false,
+      ProcessStartMode mode = ProcessStartMode.normal}) {
     throw new UnsupportedError("Process.start");
   }
 
   @patch
   static Future<ProcessResult> run(String executable, List<String> arguments,
-      {String workingDirectory,
-      Map<String, String> environment,
-      bool includeParentEnvironment: true,
-      bool runInShell: false,
-      Encoding stdoutEncoding: systemEncoding,
-      Encoding stderrEncoding: systemEncoding}) {
+      {String? workingDirectory,
+      Map<String, String>? environment,
+      bool includeParentEnvironment = true,
+      bool runInShell = false,
+      Encoding? stdoutEncoding = systemEncoding,
+      Encoding? stderrEncoding = systemEncoding}) {
     throw new UnsupportedError("Process.run");
   }
 
   @patch
   static ProcessResult runSync(String executable, List<String> arguments,
-      {String workingDirectory,
-      Map<String, String> environment,
-      bool includeParentEnvironment: true,
-      bool runInShell: false,
-      Encoding stdoutEncoding: systemEncoding,
-      Encoding stderrEncoding: systemEncoding}) {
+      {String? workingDirectory,
+      Map<String, String>? environment,
+      bool includeParentEnvironment = true,
+      bool runInShell = false,
+      Encoding? stdoutEncoding = systemEncoding,
+      Encoding? stderrEncoding = systemEncoding}) {
     throw new UnsupportedError("Process.runSync");
   }
 
@@ -386,39 +388,39 @@ class Process {
 @patch
 class InternetAddress {
   @patch
-  static InternetAddress get LOOPBACK_IP_V4 {
-    throw new UnsupportedError("InternetAddress.LOOPBACK_IP_V4");
+  static InternetAddress get loopbackIPv4 {
+    throw new UnsupportedError("InternetAddress.loopbackIPv4");
   }
 
   @patch
-  static InternetAddress get LOOPBACK_IP_V6 {
-    throw new UnsupportedError("InternetAddress.LOOPBACK_IP_V6");
+  static InternetAddress get loopbackIPv6 {
+    throw new UnsupportedError("InternetAddress.loopbackIPv6");
   }
 
   @patch
-  static InternetAddress get ANY_IP_V4 {
-    throw new UnsupportedError("InternetAddress.ANY_IP_V4");
+  static InternetAddress get anyIPv4 {
+    throw new UnsupportedError("InternetAddress.anyIPv4");
   }
 
   @patch
-  static InternetAddress get ANY_IP_V6 {
-    throw new UnsupportedError("InternetAddress.ANY_IP_V6");
+  static InternetAddress get anyIPv6 {
+    throw new UnsupportedError("InternetAddress.anyIPv6");
   }
 
   @patch
-  factory InternetAddress(String address, {InternetAddressType type}) {
+  factory InternetAddress(String address, {InternetAddressType? type}) {
     throw new UnsupportedError("InternetAddress");
   }
 
   @patch
   factory InternetAddress.fromRawAddress(Uint8List rawAddress,
-      {InternetAddressType type}) {
+      {InternetAddressType? type}) {
     throw new UnsupportedError("InternetAddress.fromRawAddress");
   }
 
   @patch
   static Future<List<InternetAddress>> lookup(String host,
-      {InternetAddressType type: InternetAddressType.any}) {
+      {InternetAddressType type = InternetAddressType.any}) {
     throw new UnsupportedError("InternetAddress.lookup");
   }
 
@@ -427,6 +429,11 @@ class InternetAddress {
       InternetAddress address, String host) {
     throw new UnsupportedError("InternetAddress._cloneWithNewHost");
   }
+
+  @patch
+  static InternetAddress? tryParse(String address) {
+    throw UnsupportedError("InternetAddress.tryParse");
+  }
 }
 
 @patch
@@ -438,9 +445,9 @@ class NetworkInterface {
 
   @patch
   static Future<List<NetworkInterface>> list(
-      {bool includeLoopback: false,
-      bool includeLinkLocal: false,
-      InternetAddressType type: InternetAddressType.any}) {
+      {bool includeLoopback = false,
+      bool includeLinkLocal = false,
+      InternetAddressType type = InternetAddressType.any}) {
     throw new UnsupportedError("NetworkInterface.list");
   }
 }
@@ -449,7 +456,7 @@ class NetworkInterface {
 class RawServerSocket {
   @patch
   static Future<RawServerSocket> bind(address, int port,
-      {int backlog: 0, bool v6Only: false, bool shared: false}) {
+      {int backlog = 0, bool v6Only = false, bool shared = false}) {
     throw new UnsupportedError("RawServerSocket.bind");
   }
 }
@@ -458,7 +465,7 @@ class RawServerSocket {
 class ServerSocket {
   @patch
   static Future<ServerSocket> _bind(address, int port,
-      {int backlog: 0, bool v6Only: false, bool shared: false}) {
+      {int backlog = 0, bool v6Only = false, bool shared = false}) {
     throw new UnsupportedError("ServerSocket.bind");
   }
 }
@@ -466,14 +473,14 @@ class ServerSocket {
 @patch
 class RawSocket {
   @patch
-  static Future<RawSocket> connect(host, int port,
-      {sourceAddress, Duration timeout}) {
+  static Future<RawSocket> connect(dynamic host, int port,
+      {dynamic sourceAddress, int sourcePort = 0, Duration? timeout}) {
     throw new UnsupportedError("RawSocket constructor");
   }
 
   @patch
-  static Future<ConnectionTask<RawSocket>> startConnect(host, int port,
-      {sourceAddress}) {
+  static Future<ConnectionTask<RawSocket>> startConnect(dynamic host, int port,
+      {dynamic sourceAddress, int sourcePort = 0}) {
     throw new UnsupportedError("RawSocket constructor");
   }
 }
@@ -481,18 +488,70 @@ class RawSocket {
 @patch
 class Socket {
   @patch
-  static Future<Socket> _connect(host, int port,
-      {sourceAddress, Duration timeout}) {
+  static Future<Socket> _connect(dynamic host, int port,
+      {dynamic sourceAddress, int sourcePort = 0, Duration? timeout}) {
     throw new UnsupportedError("Socket constructor");
   }
 
   @patch
-  static Future<ConnectionTask<Socket>> _startConnect(host, int port,
-      {sourceAddress}) {
+  static Future<ConnectionTask<Socket>> _startConnect(dynamic host, int port,
+      {dynamic sourceAddress, int sourcePort = 0}) {
     throw new UnsupportedError("Socket constructor");
   }
 }
 
+@patch
+class SocketControlMessage {
+  @patch
+  factory SocketControlMessage.fromHandles(List<ResourceHandle> handles) {
+    throw UnsupportedError("SocketControlMessage constructor");
+  }
+}
+
+@patch
+class ResourceHandle {
+  @patch
+  factory ResourceHandle.fromFile(RandomAccessFile file) {
+    throw UnsupportedError("ResourceHandle.fromFile constructor");
+  }
+
+  @patch
+  factory ResourceHandle.fromSocket(Socket socket) {
+    throw UnsupportedError("ResourceHandle.fromSocket constructor");
+  }
+
+  @patch
+  factory ResourceHandle.fromRawSocket(RawSocket rawSocket) {
+    throw UnsupportedError("ResourceHandle.fromRawSocket constructor");
+  }
+
+  @patch
+  factory ResourceHandle.fromRawDatagramSocket(
+      RawDatagramSocket rawDatagramSocket) {
+    throw UnsupportedError("ResourceHandle.fromRawDatagramSocket constructor");
+  }
+
+  @patch
+  factory ResourceHandle.fromStdin(Stdin stdin) {
+    throw UnsupportedError("ResourceHandle.fromStdin constructor");
+  }
+
+  @patch
+  factory ResourceHandle.fromStdout(Stdout stdout) {
+    throw UnsupportedError("ResourceHandle.fromStdout constructor");
+  }
+
+  @patch
+  factory ResourceHandle.fromReadPipe(ReadPipe pipe) {
+    throw UnsupportedError("ResourceHandle.fromReadPipe constructor");
+  }
+
+  @patch
+  factory ResourceHandle.fromWritePipe(WritePipe pipe) {
+    throw UnsupportedError("ResourceHandle.fromWritePipe constructor");
+  }
+}
+
 @patch
 class SecureSocket {
   @patch
@@ -504,7 +563,7 @@ class SecureSocket {
 @patch
 class RawSynchronousSocket {
   @patch
-  static RawSynchronousSocket connectSync(host, int port) {
+  static RawSynchronousSocket connectSync(dynamic host, int port) {
     throw new UnsupportedError("RawSynchronousSocket.connectSync");
   }
 }
@@ -520,7 +579,7 @@ class RawSocketOption {
 @patch
 class SecurityContext {
   @patch
-  factory SecurityContext({bool withTrustedRoots: false}) {
+  factory SecurityContext({bool withTrustedRoots = false}) {
     throw new UnsupportedError("SecurityContext constructor");
   }
 
@@ -546,8 +605,8 @@ class X509Certificate {
 @patch
 class RawDatagramSocket {
   @patch
-  static Future<RawDatagramSocket> bind(host, int port,
-      {bool reuseAddress: true, bool reusePort: false, int ttl: 1}) {
+  static Future<RawDatagramSocket> bind(dynamic host, int port,
+      {bool reuseAddress = true, bool reusePort = false, int ttl = 1}) {
     throw new UnsupportedError("RawDatagramSocket.bind");
   }
 }
@@ -608,14 +667,14 @@ class RawZLibFilter {
       int windowBits,
       int memLevel,
       int strategy,
-      List<int> dictionary,
+      List<int>? dictionary,
       bool raw) {
     throw new UnsupportedError("_newZLibDeflateFilter");
   }
 
   @patch
   static RawZLibFilter _makeZLibInflateFilter(
-      int windowBits, List<int> dictionary, bool raw) {
+      int windowBits, List<int>? dictionary, bool raw) {
     throw new UnsupportedError("_newZLibInflateFilter");
   }
 }
@@ -637,6 +696,16 @@ class Stdin {
     throw new UnsupportedError("Stdin.echoMode");
   }
 
+  @patch
+  bool get echoNewlineMode {
+    throw UnsupportedError("Stdin.echoNewlineMode");
+  }
+
+  @patch
+  void set echoNewlineMode(bool enabled) {
+    throw UnsupportedError("Stdin.echoNewlineMode");
+  }
+
   @patch
   bool get lineMode {
     throw new UnsupportedError("Stdin.lineMode");
