@@ -260,7 +260,8 @@ setNativeSubclassDispatchRecord(proto, interceptor) {
 }
 
 String constructorNameFallback(object) {
-  return JS('String', '#(#)', _constructorNameFallback, object);
+  return JS('returns:String;effects:none;depends:none', '#(#)',
+      _constructorNameFallback, object);
 }
 
 var initNativeDispatchFlag; // null or true
@@ -283,7 +284,7 @@ void initNativeDispatchContinue() {
   // tags `TAG`, if `window.TAG` is a (constructor) function, set the dispatch
   // property if the function's prototype to a dispatch record.
   var map = interceptorsByTag;
-  var tags = JS('JSMutableArray', 'Object.getOwnPropertyNames(#)', map);
+  JSArray tags = JS('JSMutableArray', 'Object.getOwnPropertyNames(#)', map);
 
   if (JS('bool', 'typeof window != "undefined"')) {
     var context = JS('=Object', 'window');
@@ -325,7 +326,7 @@ void initNativeDispatchContinue() {
 /// initialized by applying a series of hooks transformers.  Built-in hooks
 /// transformers deal with various known browser behaviours.
 ///
-/// Each hook tranformer takes a 'hooks' input which is a JavaScript object
+/// Each hook transformer takes a 'hooks' input which is a JavaScript object
 /// containing the hook functions, and returns the same or a new object with
 /// replacements.  The replacements can wrap the originals to provide alternate
 /// or modified behaviour.
