@@ -2,17 +2,22 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// All imports must be in all FFI patch files to not depend on the order
-// the patches are applied.
 import "dart:_internal" show patch;
 import 'dart:typed_data';
 import 'dart:isolate';
+import 'dart:typed_data';
 
 @pragma("vm:entry-point")
-abstract class Struct extends NativeType {}
+abstract final class _Compound extends NativeType {}
 
 @pragma("vm:entry-point")
-class _FfiStructLayout {
+abstract base class Struct extends _Compound {}
+
+@pragma("vm:entry-point")
+abstract base class Union extends _Compound {}
+
+@pragma("vm:entry-point")
+final class _FfiStructLayout {
   @pragma("vm:entry-point")
   final List<Object> fieldTypes;
 
@@ -23,7 +28,7 @@ class _FfiStructLayout {
 }
 
 @pragma("vm:entry-point")
-class _FfiInlineArray {
+final class _FfiInlineArray {
   @pragma("vm:entry-point")
   final Type elementType;
   @pragma("vm:entry-point")
