@@ -2,25 +2,25 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-import "dart:_internal" show patch;
-
 import "dart:_internal"
     show
         allocateOneByteString,
         allocateTwoByteString,
         CodeUnits,
+        ClassID,
         copyRangeFromUint8ListToOneByteString,
         doubleToIntBits,
         EfficientLengthIterable,
         FixedLengthListMixin,
+        intBitsToDouble,
         IterableElementError,
+        jsonEncode,
         ListIterator,
         Lists,
         mix64,
         POWERS_OF_TEN,
         SubListIterable,
         UnmodifiableListMixin,
-        has63BitSmis,
         makeFixedListUnmodifiable,
         makeListFixedLength,
         patch,
@@ -28,10 +28,13 @@ import "dart:_internal"
         writeIntoOneByteString,
         writeIntoTwoByteString;
 
+import "dart:_internal" as _internal show Symbol;
+
+import 'dart:_js_helper' show JS, JSSyntaxRegExp, quoteStringForRegExp;
+
 import "dart:collection"
     show
         HashMap,
-        IterableBase,
         LinkedHashMap,
         LinkedList,
         LinkedListEntry,
@@ -41,11 +44,33 @@ import "dart:collection"
         UnmodifiableMapBase,
         UnmodifiableMapView;
 
+import 'dart:convert' show Encoding, utf8;
+
 import 'dart:math' show Random;
 
-import "dart:typed_data"
-    show Endian, Uint8List, Int64List, Uint16List, Uint32List;
+import "dart:typed_data" show Uint8List, Uint16List;
 
-import 'dart:wasm';
+import 'dart:_wasm';
+
+part "bool.dart";
+part "closure.dart";
+part "double.dart";
+part "errors_patch.dart";
+part "growable_list.dart";
+part "identical_patch.dart";
+part "int.dart";
+part "list.dart";
+part "named_parameters.dart";
+part "object_patch.dart";
+part "record_patch.dart";
+part "regexp_patch.dart";
+part "stack_trace_patch.dart";
+part "stopwatch_patch.dart";
+part "string_patch.dart";
+part "type.dart";
+part "uri_patch.dart";
 
 typedef _Smi = int; // For compatibility with VM patch files
+
+String _symbolToString(Symbol s) =>
+    _internal.Symbol.getName(s as _internal.Symbol);
