@@ -2,22 +2,37 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 part of dart.io;
 
 // Read the file in blocks of size 64k.
 const int _blockSize = 64 * 1024;
 
+// The maximum number of bytes to read in a single call to `read`.
+//
+// On Windows and macOS, it is an error to call
+// `read/_read(fildes, buf, nbyte)` with `nbyte >= INT_MAX`.
+//
+// The POSIX specification states that the behavior of `read` is
+// implementation-defined if `nbyte > SSIZE_MAX`. On Linux, the `read` will
+// transfer at most 0x7ffff000 bytes and return the number of bytes actually.
+// transfered.
+//
+// A smaller value has the advantage of:
+// 1. making vm-service clients (e.g. debugger) more responsive
+// 2. reducing memory overhead (since `readInto` copies memory)
+//
+// A bigger value reduces the number of system calls.
+const int _maxReadSize = 16 * 1024 * 1024; // 16MB.
+
 class _FileStream extends Stream<List<int>> {
   // Stream controller.
-  StreamController<Uint8List> _controller;
+  late StreamController<Uint8List> _controller;
 
   // Information about the underlying file.
-  String _path;
-  RandomAccessFile _openedFile;
+  String? _path;
+  RandomAccessFile? _openedFile;
   int _position;
-  int _end;
+  int? _end;
   final Completer _closeCompleter = new Completer();
 
   // Has the stream been paused or unsubscribed?
@@ -29,20 +44,16 @@ class _FileStream extends Stream<List<int>> {
 
   bool _atEnd = false;
 
-  _FileStream(this._path, this._position, this._end) {
-    _position ??= 0;
-  }
+  _FileStream(this._path, int? position, this._end) : _position = position ?? 0;
 
   _FileStream.forStdin() : _position = 0;
 
-  StreamSubscription<Uint8List> listen(void onData(Uint8List event),
-      {Function onError, void onDone(), bool cancelOnError}) {
-    _setupController();
-    return _controller.stream.listen(onData,
-        onError: onError, onDone: onDone, cancelOnError: cancelOnError);
-  }
+  _FileStream.forRandomAccessFile(RandomAccessFile f)
+      : _position = 0,
+        _openedFile = f;
 
-  void _setupController() {
+  StreamSubscription<Uint8List> listen(void onData(Uint8List event)?,
+      {Function? onError, void onDone()?, bool? cancelOnError}) {
     _controller = new StreamController<Uint8List>(
         sync: true,
         onListen: _start,
@@ -51,6 +62,8 @@ class _FileStream extends Stream<List<int>> {
           _unsubscribed = true;
           return _closeFile();
         });
+    return _controller.stream.listen(onData,
+        onError: onError, onDone: onDone, cancelOnError: cancelOnError);
   }
 
   Future _closeFile() {
@@ -64,7 +77,7 @@ class _FileStream extends Stream<List<int>> {
       _controller.close();
     }
 
-    _openedFile.close().catchError(_controller.addError).whenComplete(done);
+    _openedFile!.close().catchError(_controller.addError).whenComplete(done);
     return _closeCompleter.future;
   }
 
@@ -77,32 +90,40 @@ class _FileStream extends Stream<List<int>> {
     }
     _readInProgress = true;
     int readBytes = _blockSize;
-    if (_end != null) {
-      readBytes = min(readBytes, _end - _position);
+    final end = _end;
+    if (end != null) {
+      readBytes = min(readBytes, end - _position);
       if (readBytes < 0) {
         _readInProgress = false;
         if (!_unsubscribed) {
-          _controller.addError(new RangeError("Bad end position: $_end"));
+          _controller.addError(new RangeError("Bad end position: $end"));
           _closeFile();
           _unsubscribed = true;
         }
         return;
       }
     }
-    _openedFile.read(readBytes).then((block) {
+    _openedFile!.read(readBytes).then((block) {
       _readInProgress = false;
       if (_unsubscribed) {
         _closeFile();
         return;
       }
       _position += block.length;
-      if (block.length < readBytes || (_end != null && _position == _end)) {
+      // read() may return less than `readBytes` if `_openFile` is a pipe or
+      // terminal or if a signal is received. Only a empty return indicates
+      // that the write side of the pipe is closed or that we are at the end
+      // of a file.
+      // See https://man7.org/linux/man-pages/man2/read.2.html
+      if (block.length == 0 || (_end != null && _position == _end)) {
         _atEnd = true;
       }
       if (!_atEnd && !_controller.isPaused) {
         _readBlock();
       }
-      _controller.add(block);
+      if (block.length > 0) {
+        _controller.add(block);
+      }
       if (_atEnd) {
         _closeFile();
       }
@@ -147,8 +168,12 @@ class _FileStream extends Stream<List<int>> {
       _closeCompleter.complete();
     }
 
-    if (_path != null) {
-      new File(_path)
+    final path = _path;
+    final openedFile = _openedFile;
+    if (openedFile != null) {
+      onOpenFile(openedFile);
+    } else if (path != null) {
+      new File(path)
           .open(mode: FileMode.read)
           .then(onOpenFile, onError: openFailed);
     } else {
@@ -161,23 +186,25 @@ class _FileStream extends Stream<List<int>> {
   }
 }
 
-class _FileStreamConsumer extends StreamConsumer<List<int>> {
-  File _file;
+class _FileStreamConsumer implements StreamConsumer<List<int>> {
+  File? _file;
   Future<RandomAccessFile> _openFuture;
 
-  _FileStreamConsumer(this._file, FileMode mode) {
-    _openFuture = _file.open(mode: mode);
-  }
+  _FileStreamConsumer(File file, FileMode mode)
+      : _file = file,
+        _openFuture = file.open(mode: mode);
 
-  _FileStreamConsumer.fromStdio(int fd) {
-    _openFuture = new Future.value(_File._openStdioSync(fd));
-  }
+  _FileStreamConsumer.fromStdio(int fd)
+      : _openFuture = new Future.value(_File._openStdioSync(fd));
 
-  Future<File> addStream(Stream<List<int>> stream) {
-    Completer<File> completer = new Completer<File>.sync();
+  _FileStreamConsumer.fromRandomAccessFile(RandomAccessFile f)
+      : _openFuture = Future.value(f);
+
+  Future<File?> addStream(Stream<List<int>> stream) {
+    Completer<File?> completer = new Completer<File?>.sync();
     _openFuture.then((openedFile) {
-      var _subscription;
-      void error(e, [StackTrace stackTrace]) {
+      late StreamSubscription<List<int>> _subscription;
+      void error(e, StackTrace stackTrace) {
         _subscription.cancel();
         openedFile.close();
         completer.completeError(e, stackTrace);
@@ -199,26 +226,23 @@ class _FileStreamConsumer extends StreamConsumer<List<int>> {
     return completer.future;
   }
 
-  Future<File> close() =>
+  Future<File?> close() =>
       _openFuture.then((openedFile) => openedFile.close()).then((_) => _file);
 }
 
 // Class for encapsulating the native implementation of files.
 class _File extends FileSystemEntity implements File {
-  String _path;
-  Uint8List _rawPath;
+  final String _path;
+  final Uint8List _rawPath;
 
-  _File(String path) {
-    ArgumentError.checkNotNull(path, 'path');
-    _path = path;
-    _rawPath = FileSystemEntity._toUtf8Array(path);
-  }
+  _File(String path)
+      : _path = _checkNotNull(path, "path"),
+        _rawPath = FileSystemEntity._toUtf8Array(path);
 
-  _File.fromRawPath(Uint8List rawPath) {
-    ArgumentError.checkNotNull(rawPath, 'rawPath');
-    _rawPath = FileSystemEntity._toNullTerminatedUtf8Array(rawPath);
-    _path = FileSystemEntity._toStringFromUtf8Array(rawPath);
-  }
+  _File.fromRawPath(Uint8List rawPath)
+      : _rawPath = FileSystemEntity._toNullTerminatedUtf8Array(
+            _checkNotNull(rawPath, "rawPath")),
+        _path = FileSystemEntity._toStringFromUtf8Array(rawPath);
 
   String get path => _path;
 
@@ -229,7 +253,7 @@ class _File extends FileSystemEntity implements File {
   // with it.
   static int _namespacePointer() => _Namespace._namespacePointer;
 
-  static Future _dispatchWithNamespace(int request, List data) {
+  static Future<Object?> _dispatchWithNamespace(int request, List data) {
     data[0] = _namespacePointer();
     return _IOService._dispatch(request, data);
   }
@@ -237,10 +261,8 @@ class _File extends FileSystemEntity implements File {
   Future<bool> exists() {
     return _dispatchWithNamespace(_IOService.fileExists, [null, _rawPath])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "Cannot check existence", path);
-      }
-      return response;
+      _checkForErrorResponse(response, "Cannot check existence", path);
+      return response as bool;
     });
   }
 
@@ -254,44 +276,43 @@ class _File extends FileSystemEntity implements File {
 
   File get absolute => new File(_absolutePath);
 
-  Future<File> create({bool recursive: false}) {
+  Future<File> create({bool recursive = false, bool exclusive = false}) {
     var result =
         recursive ? parent.create(recursive: true) : new Future.value(null);
     return result
-        .then((_) =>
-            _dispatchWithNamespace(_IOService.fileCreate, [null, _rawPath]))
+        .then((_) => _dispatchWithNamespace(
+            _IOService.fileCreate, [null, _rawPath, exclusive]))
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "Cannot create file", path);
-      }
+      _checkForErrorResponse(response, "Cannot create file", path);
       return this;
     });
   }
 
-  external static _create(_Namespace namespace, Uint8List rawPath);
+  external static _create(
+      _Namespace namespace, Uint8List rawPath, bool exclusive);
 
   external static _createLink(
       _Namespace namespace, Uint8List rawPath, String target);
 
+  external static List<dynamic> _createPipe(_Namespace namespace);
+
   external static _linkTarget(_Namespace namespace, Uint8List rawPath);
 
-  void createSync({bool recursive: false}) {
+  void createSync({bool recursive = false, bool exclusive = false}) {
     if (recursive) {
       parent.createSync(recursive: true);
     }
-    var result = _create(_Namespace._namespace, _rawPath);
+    var result = _create(_Namespace._namespace, _rawPath, exclusive);
     throwIfError(result, "Cannot create file", path);
   }
 
-  Future<File> _delete({bool recursive: false}) {
+  Future<File> _delete({bool recursive = false}) {
     if (recursive) {
       return new Directory(path).delete(recursive: true).then((_) => this);
     }
     return _dispatchWithNamespace(_IOService.fileDelete, [null, _rawPath])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "Cannot delete file", path);
-      }
+      _checkForErrorResponse(response, "Cannot delete file", path);
       return this;
     });
   }
@@ -300,7 +321,7 @@ class _File extends FileSystemEntity implements File {
 
   external static _deleteLinkNative(_Namespace namespace, Uint8List rawPath);
 
-  void _deleteSync({bool recursive: false}) {
+  void _deleteSync({bool recursive = false}) {
     if (recursive) {
       return new Directory.fromRawPath(_rawPath).deleteSync(recursive: true);
     }
@@ -311,10 +332,8 @@ class _File extends FileSystemEntity implements File {
   Future<File> rename(String newPath) {
     return _dispatchWithNamespace(
         _IOService.fileRename, [null, _rawPath, newPath]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(
-            response, "Cannot rename file to '$newPath'", path);
-      }
+      _checkForErrorResponse(
+          response, "Cannot rename file to '$newPath'", path);
       return new File(newPath);
     });
   }
@@ -334,10 +353,7 @@ class _File extends FileSystemEntity implements File {
   Future<File> copy(String newPath) {
     return _dispatchWithNamespace(
         _IOService.fileCopy, [null, _rawPath, newPath]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(
-            response, "Cannot copy file to '$newPath'", path);
-      }
+      _checkForErrorResponse(response, "Cannot copy file to '$newPath'", path);
       return new File(newPath);
     });
   }
@@ -351,7 +367,7 @@ class _File extends FileSystemEntity implements File {
     return new File(newPath);
   }
 
-  Future<RandomAccessFile> open({FileMode mode: FileMode.read}) {
+  Future<RandomAccessFile> open({FileMode mode = FileMode.read}) {
     if (mode != FileMode.read &&
         mode != FileMode.write &&
         mode != FileMode.append &&
@@ -362,21 +378,16 @@ class _File extends FileSystemEntity implements File {
     }
     return _dispatchWithNamespace(
         _IOService.fileOpen, [null, _rawPath, mode._mode]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "Cannot open file", path);
-      }
-      return new _RandomAccessFile(response, path);
+      _checkForErrorResponse(response, "Cannot open file", path);
+      return _RandomAccessFile(response as int, path);
     });
   }
 
   Future<int> length() {
     return _dispatchWithNamespace(
         _IOService.fileLengthFromPath, [null, _rawPath]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(
-            response, "Cannot retrieve length of file", path);
-      }
-      return response;
+      _checkForErrorResponse(response, "Cannot retrieve length of file", path);
+      return response as int;
     });
   }
 
@@ -391,11 +402,8 @@ class _File extends FileSystemEntity implements File {
   Future<DateTime> lastAccessed() {
     return _dispatchWithNamespace(_IOService.fileLastAccessed, [null, _rawPath])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(
-            response, "Cannot retrieve access time", path);
-      }
-      return new DateTime.fromMillisecondsSinceEpoch(response);
+      _checkForErrorResponse(response, "Cannot retrieve access time", path);
+      return DateTime.fromMillisecondsSinceEpoch(response as int);
     });
   }
 
@@ -412,9 +420,7 @@ class _File extends FileSystemEntity implements File {
     return _dispatchWithNamespace(
             _IOService.fileSetLastAccessed, [null, _rawPath, millis])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "Cannot set access time", path);
-      }
+      _checkForErrorResponse(response, "Cannot set access time", path);
       return null;
     });
   }
@@ -434,11 +440,9 @@ class _File extends FileSystemEntity implements File {
   Future<DateTime> lastModified() {
     return _dispatchWithNamespace(_IOService.fileLastModified, [null, _rawPath])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(
-            response, "Cannot retrieve modification time", path);
-      }
-      return new DateTime.fromMillisecondsSinceEpoch(response);
+      _checkForErrorResponse(
+          response, "Cannot retrieve modification time", path);
+      return DateTime.fromMillisecondsSinceEpoch(response as int);
     });
   }
 
@@ -455,10 +459,7 @@ class _File extends FileSystemEntity implements File {
     return _dispatchWithNamespace(
             _IOService.fileSetLastModified, [null, _rawPath, millis])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(
-            response, "Cannot set modification time", path);
-      }
+      _checkForErrorResponse(response, "Cannot set modification time", path);
       return null;
     });
   }
@@ -477,7 +478,7 @@ class _File extends FileSystemEntity implements File {
 
   external static _open(_Namespace namespace, Uint8List rawPath, int mode);
 
-  RandomAccessFile openSync({FileMode mode: FileMode.read}) {
+  RandomAccessFile openSync({FileMode mode = FileMode.read}) {
     if (mode != FileMode.read &&
         mode != FileMode.write &&
         mode != FileMode.append &&
@@ -500,11 +501,11 @@ class _File extends FileSystemEntity implements File {
     return new _RandomAccessFile(id, "");
   }
 
-  Stream<List<int>> openRead([int start, int end]) {
+  Stream<List<int>> openRead([int? start, int? end]) {
     return new _FileStream(path, start, end);
   }
 
-  IOSink openWrite({FileMode mode: FileMode.write, Encoding encoding: utf8}) {
+  IOSink openWrite({FileMode mode = FileMode.write, Encoding encoding = utf8}) {
     if (mode != FileMode.write &&
         mode != FileMode.append &&
         mode != FileMode.writeOnly &&
@@ -516,7 +517,7 @@ class _File extends FileSystemEntity implements File {
   }
 
   Future<Uint8List> readAsBytes() {
-    Future<Uint8List> readDataChunked(RandomAccessFile file) {
+    Future<Uint8List> readUnsized(RandomAccessFile file) {
       var builder = new BytesBuilder(copy: false);
       var completer = new Completer<Uint8List>();
       void read() {
@@ -534,13 +535,37 @@ class _File extends FileSystemEntity implements File {
       return completer.future;
     }
 
+    Future<Uint8List> readSized(RandomAccessFile file, int length) {
+      var data = Uint8List(length);
+      var offset = 0;
+      var completer = new Completer<Uint8List>();
+      void read() {
+        file.readInto(data, offset, min(offset + _maxReadSize, length)).then(
+            (readSize) {
+          if (readSize > 0) {
+            offset += readSize;
+            read();
+          } else {
+            assert(readSize == 0);
+            if (offset < length) {
+              data = Uint8List.sublistView(data, 0, offset);
+            }
+            completer.complete(data);
+          }
+        }, onError: completer.completeError);
+      }
+
+      read();
+      return completer.future;
+    }
+
     return open().then((file) {
       return file.length().then((length) {
         if (length == 0) {
           // May be character device, try to read it in chunks.
-          return readDataChunked(file);
+          return readUnsized(file);
         }
-        return file.read(length);
+        return readSized(file, length);
       }).whenComplete(file.close);
     });
   }
@@ -555,11 +580,27 @@ class _File extends FileSystemEntity implements File {
         var builder = new BytesBuilder(copy: false);
         do {
           data = opened.readSync(_blockSize);
-          if (data.length > 0) builder.add(data);
+          if (data.length > 0) {
+            builder.add(data);
+          }
         } while (data.length > 0);
         data = builder.takeBytes();
       } else {
-        data = opened.readSync(length);
+        data = Uint8List(length);
+        var offset = 0;
+
+        while (offset < length) {
+          final readSize = opened.readIntoSync(
+              data, offset, min(offset + _maxReadSize, length));
+          if (readSize == 0) {
+            break;
+          }
+          offset += readSize;
+        }
+
+        if (offset < length) {
+          data = Uint8List.view(data.buffer, 0, offset);
+        }
       }
       return data;
     } finally {
@@ -576,31 +617,20 @@ class _File extends FileSystemEntity implements File {
     }
   }
 
-  Future<String> readAsString({Encoding encoding: utf8}) {
-    // TODO(dart:io): If the change in async semantics to run synchronously
-    // until await lands, this is as efficient as
-    // return _tryDecode(await readAsBytes(), encoding);
-    var stack = StackTrace.current;
-    return readAsBytes().then((bytes) {
-      try {
-        return _tryDecode(bytes, encoding);
-      } catch (e) {
-        return new Future.error(e, stack);
-      }
-    });
-  }
+  Future<String> readAsString({Encoding encoding = utf8}) async =>
+      _tryDecode(await readAsBytes(), encoding);
 
-  String readAsStringSync({Encoding encoding: utf8}) =>
+  String readAsStringSync({Encoding encoding = utf8}) =>
       _tryDecode(readAsBytesSync(), encoding);
 
-  Future<List<String>> readAsLines({Encoding encoding: utf8}) =>
+  Future<List<String>> readAsLines({Encoding encoding = utf8}) =>
       readAsString(encoding: encoding).then(const LineSplitter().convert);
 
-  List<String> readAsLinesSync({Encoding encoding: utf8}) =>
+  List<String> readAsLinesSync({Encoding encoding = utf8}) =>
       const LineSplitter().convert(readAsStringSync(encoding: encoding));
 
   Future<File> writeAsBytes(List<int> bytes,
-      {FileMode mode: FileMode.write, bool flush: false}) {
+      {FileMode mode = FileMode.write, bool flush = false}) {
     return open(mode: mode).then((file) {
       return file.writeFrom(bytes, 0, bytes.length).then<File>((_) {
         if (flush) return file.flush().then((_) => this);
@@ -610,7 +640,7 @@ class _File extends FileSystemEntity implements File {
   }
 
   void writeAsBytesSync(List<int> bytes,
-      {FileMode mode: FileMode.write, bool flush: false}) {
+      {FileMode mode = FileMode.write, bool flush = false}) {
     RandomAccessFile opened = openSync(mode: mode);
     try {
       opened.writeFromSync(bytes, 0, bytes.length);
@@ -621,9 +651,9 @@ class _File extends FileSystemEntity implements File {
   }
 
   Future<File> writeAsString(String contents,
-      {FileMode mode: FileMode.write,
-      Encoding encoding: utf8,
-      bool flush: false}) {
+      {FileMode mode = FileMode.write,
+      Encoding encoding = utf8,
+      bool flush = false}) {
     try {
       return writeAsBytes(encoding.encode(contents), mode: mode, flush: flush);
     } catch (e) {
@@ -632,9 +662,9 @@ class _File extends FileSystemEntity implements File {
   }
 
   void writeAsStringSync(String contents,
-      {FileMode mode: FileMode.write,
-      Encoding encoding: utf8,
-      bool flush: false}) {
+      {FileMode mode = FileMode.write,
+      Encoding encoding = utf8,
+      bool flush = false}) {
     writeAsBytesSync(encoding.encode(contents), mode: mode, flush: flush);
   }
 
@@ -642,21 +672,28 @@ class _File extends FileSystemEntity implements File {
 
   static throwIfError(Object result, String msg, String path) {
     if (result is OSError) {
-      throw new FileSystemException(msg, path, result);
+      throw FileSystemException._fromOSError(result, msg, path);
     }
   }
+
+  // TODO(40614): Remove once non-nullability is sound.
+  static T _checkNotNull<T>(T t, String name) {
+    ArgumentError.checkNotNull(t, name);
+    return t;
+  }
 }
 
 abstract class _RandomAccessFileOps {
   external factory _RandomAccessFileOps(int pointer);
 
   int getPointer();
+  int get fd;
   int close();
   readByte();
   read(int bytes);
-  readInto(List<int> buffer, int start, int end);
+  readInto(List<int> buffer, int start, int? end);
   writeByte(int value);
-  writeFrom(List<int> buffer, int start, int end);
+  writeFrom(List<int> buffer, int start, int? end);
   position();
   setPosition(int position);
   truncate(int length);
@@ -665,26 +702,27 @@ abstract class _RandomAccessFileOps {
   lock(int lock, int start, int end);
 }
 
+@pragma("vm:entry-point")
 class _RandomAccessFile implements RandomAccessFile {
   static bool _connectedResourceHandler = false;
 
   final String path;
 
   bool _asyncDispatched = false;
-  SendPort _fileService;
 
-  _FileResourceInfo _resourceInfo;
+  late _FileResourceInfo _resourceInfo;
   _RandomAccessFileOps _ops;
 
-  _RandomAccessFile(int pointer, this.path) {
-    _ops = new _RandomAccessFileOps(pointer);
+  @pragma("vm:entry-point")
+  _RandomAccessFile(int pointer, this.path)
+      : _ops = new _RandomAccessFileOps(pointer) {
     _resourceInfo = new _FileResourceInfo(this);
     _maybeConnectHandler();
   }
 
   void _maybePerformCleanup() {
     if (closed) {
-      _FileResourceInfo.FileClosed(_resourceInfo);
+      _FileResourceInfo.fileClosed(_resourceInfo);
     }
   }
 
@@ -696,8 +734,8 @@ class _RandomAccessFile implements RandomAccessFile {
       // open.
       registerExtension(
           'ext.dart.io.getOpenFiles', _FileResourceInfo.getOpenFiles);
-      registerExtension(
-          'ext.dart.io.getFileByID', _FileResourceInfo.getFileInfoMapByID);
+      registerExtension('ext.dart.io.getOpenFileById',
+          _FileResourceInfo.getOpenFileInfoMapByID);
       _connectedResourceHandler = true;
     }
   }
@@ -725,11 +763,9 @@ class _RandomAccessFile implements RandomAccessFile {
 
   Future<int> readByte() {
     return _dispatch(_IOService.fileReadByte, [null]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "readByte failed", path);
-      }
+      _checkForErrorResponse(response, "readByte failed", path);
       _resourceInfo.addRead(1);
-      return response;
+      return response as int;
     });
   }
 
@@ -744,58 +780,51 @@ class _RandomAccessFile implements RandomAccessFile {
   }
 
   Future<Uint8List> read(int bytes) {
-    ArgumentError.checkNotNull(bytes, 'bytes');
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(bytes, "bytes");
     return _dispatch(_IOService.fileRead, [null, bytes]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "read failed", path);
-      }
-      _resourceInfo.addRead(response[1].length);
-      Uint8List result = response[1];
+      _checkForErrorResponse(response, "read failed", path);
+      var result = (response as List<Object?>)[1] as Uint8List;
+      _resourceInfo.addRead(result.length);
       return result;
     });
   }
 
   Uint8List readSync(int bytes) {
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(bytes, "bytes");
     _checkAvailable();
-    ArgumentError.checkNotNull(bytes, 'bytes');
     var result = _ops.read(bytes);
-    if (result is OSError) {
-      throw new FileSystemException("readSync failed", path, result);
+    if (result is! Uint8List) {
+      throw new FileSystemException("readSync failed", path, result as OSError);
     }
     _resourceInfo.addRead(result.length);
     return result;
   }
 
-  Future<int> readInto(List<int> buffer, [int start = 0, int end]) {
-    if ((buffer is! List) ||
-        ((start != null) && (start is! int)) ||
-        ((end != null) && (end is! int))) {
-      throw new ArgumentError();
-    }
+  Future<int> readInto(List<int> buffer, [int start = 0, int? end]) {
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(buffer, "buffer");
     end = RangeError.checkValidRange(start, end, buffer.length);
     if (end == start) {
       return new Future.value(0);
     }
     int length = end - start;
     return _dispatch(_IOService.fileReadInto, [null, length]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "readInto failed", path);
-      }
-      int read = response[1];
-      List<int> data = response[2];
+      _checkForErrorResponse(response, "readInto failed", path);
+      var responseList = response as List<Object?>;
+      var read = responseList[1] as int;
+      var data = responseList[2] as List<int>;
       buffer.setRange(start, start + read, data);
       _resourceInfo.addRead(read);
       return read;
     });
   }
 
-  int readIntoSync(List<int> buffer, [int start = 0, int end]) {
+  int readIntoSync(List<int> buffer, [int start = 0, int? end]) {
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(buffer, "buffer");
     _checkAvailable();
-    if ((buffer is! List) ||
-        ((start != null) && (start is! int)) ||
-        ((end != null) && (end is! int))) {
-      throw new ArgumentError();
-    }
     end = RangeError.checkValidRange(start, end, buffer.length);
     if (end == start) {
       return 0;
@@ -809,11 +838,10 @@ class _RandomAccessFile implements RandomAccessFile {
   }
 
   Future<RandomAccessFile> writeByte(int value) {
-    ArgumentError.checkNotNull(value, 'value');
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(value, "value");
     return _dispatch(_IOService.fileWriteByte, [null, value]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "writeByte failed", path);
-      }
+      _checkForErrorResponse(response, "writeByte failed", path);
       _resourceInfo.addWrite(1);
       return this;
     });
@@ -821,7 +849,8 @@ class _RandomAccessFile implements RandomAccessFile {
 
   int writeByteSync(int value) {
     _checkAvailable();
-    ArgumentError.checkNotNull(value, 'value');
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(value, "value");
     var result = _ops.writeByte(value);
     if (result is OSError) {
       throw new FileSystemException("writeByte failed", path, result);
@@ -831,12 +860,10 @@ class _RandomAccessFile implements RandomAccessFile {
   }
 
   Future<RandomAccessFile> writeFrom(List<int> buffer,
-      [int start = 0, int end]) {
-    if ((buffer is! List) ||
-        ((start != null) && (start is! int)) ||
-        ((end != null) && (end is! int))) {
-      throw new ArgumentError("Invalid arguments to writeFrom");
-    }
+      [int start = 0, int? end]) {
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(buffer, "buffer");
+    ArgumentError.checkNotNull(start, "start");
     end = RangeError.checkValidRange(start, end, buffer.length);
     if (end == start) {
       return new Future.value(this);
@@ -848,27 +875,23 @@ class _RandomAccessFile implements RandomAccessFile {
       return new Future.error(e);
     }
 
-    List request = new List(4);
+    List request = new List<dynamic>.filled(4, null);
     request[0] = null;
     request[1] = result.buffer;
     request[2] = result.start;
     request[3] = end - (start - result.start);
     return _dispatch(_IOService.fileWriteFrom, request).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "writeFrom failed", path);
-      }
-      _resourceInfo.addWrite(end - (start - result.start));
+      _checkForErrorResponse(response, "writeFrom failed", path);
+      _resourceInfo.addWrite(end! - (start - result.start));
       return this;
     });
   }
 
-  void writeFromSync(List<int> buffer, [int start = 0, int end]) {
+  void writeFromSync(List<int> buffer, [int start = 0, int? end]) {
     _checkAvailable();
-    if ((buffer is! List) ||
-        ((start != null) && (start is! int)) ||
-        ((end != null) && (end is! int))) {
-      throw new ArgumentError("Invalid arguments to writeFromSync");
-    }
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(buffer, "buffer");
+    ArgumentError.checkNotNull(start, "start");
     end = RangeError.checkValidRange(start, end, buffer.length);
     if (end == start) {
       return;
@@ -884,24 +907,24 @@ class _RandomAccessFile implements RandomAccessFile {
   }
 
   Future<RandomAccessFile> writeString(String string,
-      {Encoding encoding: utf8}) {
-    ArgumentError.checkNotNull(encoding, 'encoding');
+      {Encoding encoding = utf8}) {
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(encoding, "encoding");
     var data = encoding.encode(string);
     return writeFrom(data, 0, data.length);
   }
 
-  void writeStringSync(String string, {Encoding encoding: utf8}) {
-    ArgumentError.checkNotNull(encoding, 'encoding');
+  void writeStringSync(String string, {Encoding encoding = utf8}) {
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(encoding, "encoding");
     var data = encoding.encode(string);
     writeFromSync(data, 0, data.length);
   }
 
   Future<int> position() {
     return _dispatch(_IOService.filePosition, [null]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "position failed", path);
-      }
-      return response;
+      _checkForErrorResponse(response, "position failed", path);
+      return response as int;
     });
   }
 
@@ -917,9 +940,7 @@ class _RandomAccessFile implements RandomAccessFile {
   Future<RandomAccessFile> setPosition(int position) {
     return _dispatch(_IOService.fileSetPosition, [null, position])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "setPosition failed", path);
-      }
+      _checkForErrorResponse(response, "setPosition failed", path);
       return this;
     });
   }
@@ -934,9 +955,7 @@ class _RandomAccessFile implements RandomAccessFile {
 
   Future<RandomAccessFile> truncate(int length) {
     return _dispatch(_IOService.fileTruncate, [null, length]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "truncate failed", path);
-      }
+      _checkForErrorResponse(response, "truncate failed", path);
       return this;
     });
   }
@@ -951,10 +970,8 @@ class _RandomAccessFile implements RandomAccessFile {
 
   Future<int> length() {
     return _dispatch(_IOService.fileLength, [null]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "length failed", path);
-      }
-      return response;
+      _checkForErrorResponse(response, "length failed", path);
+      return response as int;
     });
   }
 
@@ -969,9 +986,7 @@ class _RandomAccessFile implements RandomAccessFile {
 
   Future<RandomAccessFile> flush() {
     return _dispatch(_IOService.fileFlush, [null]).then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, "flush failed", path);
-      }
+      _checkForErrorResponse(response, "flush failed", path);
       return this;
     });
   }
@@ -994,34 +1009,31 @@ class _RandomAccessFile implements RandomAccessFile {
 
   Future<RandomAccessFile> lock(
       [FileLock mode = FileLock.exclusive, int start = 0, int end = -1]) {
-    if ((mode is! FileLock) || (start is! int) || (end is! int)) {
-      throw new ArgumentError();
-    }
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(mode, "mode");
+    ArgumentError.checkNotNull(start, "start");
+    ArgumentError.checkNotNull(end, "end");
     if ((start < 0) || (end < -1) || ((end != -1) && (start >= end))) {
       throw new ArgumentError();
     }
     int lock = _fileLockValue(mode);
     return _dispatch(_IOService.fileLock, [null, lock, start, end])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, 'lock failed', path);
-      }
+      _checkForErrorResponse(response, 'lock failed', path);
       return this;
     });
   }
 
   Future<RandomAccessFile> unlock([int start = 0, int end = -1]) {
-    if ((start is! int) || (end is! int)) {
-      throw new ArgumentError();
-    }
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(start, "start");
+    ArgumentError.checkNotNull(end, "end");
     if (start == end) {
       throw new ArgumentError();
     }
     return _dispatch(_IOService.fileLock, [null, lockUnlock, start, end])
         .then((response) {
-      if (_isErrorResponse(response)) {
-        throw _exceptionFromResponse(response, 'unlock failed', path);
-      }
+      _checkForErrorResponse(response, 'unlock failed', path);
       return this;
     });
   }
@@ -1029,9 +1041,10 @@ class _RandomAccessFile implements RandomAccessFile {
   void lockSync(
       [FileLock mode = FileLock.exclusive, int start = 0, int end = -1]) {
     _checkAvailable();
-    if ((mode is! FileLock) || (start is! int) || (end is! int)) {
-      throw new ArgumentError();
-    }
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(mode, "mode");
+    ArgumentError.checkNotNull(start, "start");
+    ArgumentError.checkNotNull(end, "end");
     if ((start < 0) || (end < -1) || ((end != -1) && (start >= end))) {
       throw new ArgumentError();
     }
@@ -1044,9 +1057,9 @@ class _RandomAccessFile implements RandomAccessFile {
 
   void unlockSync([int start = 0, int end = -1]) {
     _checkAvailable();
-    if ((start is! int) || (end is! int)) {
-      throw new ArgumentError();
-    }
+    // TODO(40614): Remove once non-nullability is sound.
+    ArgumentError.checkNotNull(start, "start");
+    ArgumentError.checkNotNull(end, "end");
     if (start == end) {
       throw new ArgumentError();
     }
@@ -1058,6 +1071,8 @@ class _RandomAccessFile implements RandomAccessFile {
 
   bool closed = false;
 
+  int get fd => _ops.fd;
+
   // WARNING:
   // Calling this function will increase the reference count on the native
   // object that implements the file operations. It should only be called to
@@ -1065,7 +1080,7 @@ class _RandomAccessFile implements RandomAccessFile {
   // count when it is finished with it.
   int _pointer() => _ops.getPointer();
 
-  Future _dispatch(int request, List data, {bool markClosed: false}) {
+  Future<Object?> _dispatch(int request, List data, {bool markClosed = false}) {
     if (closed) {
       return new Future.error(new FileSystemException("File closed", path));
     }
@@ -1095,3 +1110,43 @@ class _RandomAccessFile implements RandomAccessFile {
     }
   }
 }
+
+class _ReadPipe extends _FileStream implements ReadPipe {
+  _ReadPipe(RandomAccessFile file) : super.forRandomAccessFile(file);
+}
+
+class _WritePipe extends _IOSinkImpl implements WritePipe {
+  RandomAccessFile _file;
+  _WritePipe(file)
+      : this._file = file,
+        super(_FileStreamConsumer.fromRandomAccessFile(file), utf8);
+}
+
+class _Pipe implements Pipe {
+  final ReadPipe _readPipe;
+  final WritePipe _writePipe;
+
+  ReadPipe get read => _readPipe;
+  WritePipe get write => _writePipe;
+
+  _Pipe(this._readPipe, this._writePipe);
+
+  static Future<_Pipe> create() {
+    final completer = Completer<_Pipe>.sync();
+
+    _File._dispatchWithNamespace(_IOService.fileCreatePipe, [null])
+        .then((response) {
+      final filePointers = (response as List).cast<int>();
+      completer.complete(_Pipe(
+          _ReadPipe(_RandomAccessFile(filePointers[0], '')),
+          _WritePipe(_RandomAccessFile(filePointers[1], ''))));
+    });
+    return completer.future;
+  }
+
+  factory _Pipe.createSync() {
+    final filePointers = _File._createPipe(_Namespace._namespace);
+    return _Pipe(_ReadPipe(_RandomAccessFile(filePointers[0] as int, '')),
+        _WritePipe(_RandomAccessFile(filePointers[1] as int, '')));
+  }
+}
